// Copyright (C) 2006-2017 Tuma Solutions, LLC
// Process Dashboard - Data Automation Tool for high-maturity processes
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// Additional permissions also apply; see the README-license.txt
// file in the project root directory for more information.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
//
// The author(s) may be contacted at:
//     processdash@tuma-solutions.com
//     processdash-devel@lists.sourceforge.net

package net.sourceforge.processdash.net.cms;

/*
 * Build note: this class uses many classes that are autogenerated during the
 * ant build process.  If your IDE or compiler is giving you error messages
 * about missing classes, you need to run ant!  See the file "README-build.txt"
 * in the root directory of this project for more information.
 */

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sourceforge.processdash.data.DataContext;
import net.sourceforge.processdash.i18n.Resources;
import net.sourceforge.processdash.tool.perm.PermissionsManager;
import net.sourceforge.processdash.ui.snippet.SnippetDefinitionManager;
import net.sourceforge.processdash.ui.web.reports.analysis.AnalysisPage;
import net.sourceforge.processdash.util.HTMLUtils;

/** Abstract base class for renderers that generate a single page of content.
 */
public abstract class AbstractPageAssembler implements PageAssembler,
        CMSSnippetEnvironment, Needs.Environment, Needs.Parameters,
        Needs.Prefix, Needs.Data {

    protected static final String AUTO_HEADER_INSTANCE_ID = "auto";
    protected static final Resources resources = Resources
            .getDashBundle("CMS.Snippet");

    protected static final String HTML_STRICT_DOCTYPE =
            HTMLUtils.HTML_STRICT_DOCTYPE;
    protected static final String HTML_TRANSITIONAL_DOCTYPE =
            HTMLUtils.HTML_TRANSITIONAL_DOCTYPE;
    protected static final String HTML_FRAMESET_DOCTYPE =
            HTMLUtils.HTML_FRAMESET_DOCTYPE;

    // collect information, via the Needs interfaces.

    protected Map environment;

    protected Map parameters;

    protected String prefix;

    protected DataContext dataContext;

    public void setData(DataContext dataContext) {
        this.dataContext = dataContext;
    }

    public void setEnvironment(Map environment) {
        this.environment = environment;
    }

    public void setParameters(Map parameters) {
        this.parameters = parameters;
    }

    public void setPrefix(String prefix) {
        this.prefix = prefix;
    }


    // methods that subclasses can override.

    protected abstract void writePage(Writer out, Set headerItems,
            PageContentTO page) throws IOException;

    private boolean shouldClobberForms() {
        return false;
    }

    protected void setSnippetNamespace(SnippetInstanceTO snippet,
            String namespace) {
        snippet.setNamespace(namespace);
    }

    protected void addPageSpecificHeaderItems(Set headerItems) {
        addStyleSheet(headerItems, "/style.css");
    }


    /**  Handle the default tasks associated with rendering a page.
     */
    public void service(Writer out, PageContentTO page) throws IOException {

        addPageSpecificParameters(parameters, page);

        SnippetInvoker invoker = new SnippetInvoker(environment, parameters,
                prefix, dataContext);
        String selfURI = CmsContentDispatcher.getSimpleSelfUri(environment,
                false);
        Set headerItems = new LinkedHashSet();
        addPageSpecificHeaderItems(headerItems);

        maybeAddHeaderSnippet(page, invoker);

        int num = 0;
        for (Iterator i = page.getSnippets().iterator(); i.hasNext();) {
            SnippetInstanceTO snip = (SnippetInstanceTO) i.next();
            setSnippetNamespace(snip, "snip" + (num++) + "_");

            if (shouldInvokeSnippet(snip)) {
                try {
                    String snipContent = invoker.invoke(snip);
                    if (snipContent != null) {
                        SnippetHtmlPostprocessor post =    // see comment at top of file
                            new SnippetHtmlPostprocessor(
                                snip.getNamespace(), selfURI, snip.getUri(),
                                shouldClobberForms(), snipContent);
                        headerItems.addAll(post.getHeaderItems());
                        snip.setGeneratedContent(post.getResults());
                    } else {
                        snip.setGeneratedContent(null);
                    }
                } catch (IOException ioe) {
                    snip.setGeneratedContent(null);
                }
            } else {
                snip.setStatus(SnippetInvoker.STATUS_NOT_RUN);
                snip.setGeneratedContent(null);
                invoker.test(snip);
            }
        }

        beforeWritePage(headerItems, page);

        writePage(out, headerItems, page);
    }

    protected void maybeAddHeaderSnippet(PageContentTO page,
            SnippetInvoker invoker) {
        if (!containsHeaderSnippet(page, invoker)) {
            SnippetInstanceTO headerSnip = getDefaultHeaderSnippet();
            headerSnip.setInstanceID(AUTO_HEADER_INSTANCE_ID);
            List snippets = page.getSnippets();
            if (snippets == null)
                page.setSnippets(snippets = new ArrayList());
            snippets.add(0, headerSnip);
        }
    }
    private boolean containsHeaderSnippet(PageContentTO page,
            SnippetInvoker invoker) {
        Iterator i = page.getHeaderSnippets();
        while (i.hasNext()) {
            SnippetInstanceTO snip = (SnippetInstanceTO) i.next();
            if (snip.getDefinition() != null) {
                String mode = (String) parameters.get("mode");
                if (mode == null || "view".equals(mode)
                        || snip.getDefinition().getModes().contains(mode))
                    if (invoker.test(snip))
                        return true;
            }
        }
        return false;
    }
    protected SnippetInstanceTO getDefaultHeaderSnippet() {
        SnippetInstanceTO snip = new SnippetInstanceTO();
        String snipID = getDefaultHeaderSnippetID();
        snip.setSnippetID(snipID);
        snip.setDefinition(SnippetDefinitionManager.getSnippet(snipID));
        snip.setPageRegion(PageContentTO.REGION_HEADER);
        return snip;
    }
    protected String getDefaultHeaderSnippetID() {
        return "pdash.pageHeading";
    }

    protected void addPageSpecificParameters(Map params, PageContentTO page) {
        params.put(CMS_PAGE_FILENAME, getPageFilename());
        params.put(CMS_PAGE_TITLE, page.getPageTitle());
        params.put(CMS_LOCALIZED_PREFIX, AnalysisPage.localizePrefix(prefix));

        String pageUri = (String) environment.get("REQUEST_URI");
        pageUri = HTMLUtils.removeParam(pageUri, "mode");
        pageUri = HTMLUtils.removeParam(pageUri, "EXPORT");
        params.put(CURRENT_FRAME_URI, pageUri);
        params.put(FULL_PAGE_URI, pageUri);
        params.put(FULL_PAGE_TARGET, "_top");
    }

    private String getPageFilename() {
        String result = (String) environment.get("SCRIPT_NAME");
        result = result.substring(CmsContentDispatcher.CMS_URI_PREFIX.length());
        result = HTMLUtils.urlDecode(result);
        return result;
    }

    /** Return true if the SnippetInvoker should be run on the given snippet.
     * 
     * Page assemblers which do not display all of the snippets on a page may
     * choose to override this, returning false for snippets that will not
     * be displayed, to speed the rendering of pages.
     */
    protected boolean shouldInvokeSnippet(SnippetInstanceTO snip) {
        switch (snip.getPageRegion()) {
        case PageContentTO.REGION_HEADER:
            return shouldInvokerHeaderSnippet(snip);
        case PageContentTO.REGION_CONTENT:
            return shouldInvokeContentSnippet(snip);
        case PageContentTO.REGION_FOOTER:
            return shouldInvokeFooterSnippet(snip);
        default:
            return true;
        }
    }
    protected boolean shouldInvokerHeaderSnippet(SnippetInstanceTO snip) {
        return true;
    }

    protected boolean shouldInvokeContentSnippet(SnippetInstanceTO snip) {
        return true;
    }

    protected boolean shouldInvokeFooterSnippet(SnippetInstanceTO snip) {
        return true;
    }

    protected void beforeWritePage(Set headerItems, PageContentTO page) {
        // no-op: subclasses can override
    }

    protected void writeHead(Writer out, Set headerItems, PageContentTO page) throws IOException {
        out.write("<head>\n");

        String pageTitle = getPageTitle(page);
        out.write("<title>");
        out.write(pageTitle);
        out.write("</title>\n");

        out.write("\n");

        for (Iterator i = headerItems.iterator(); i.hasNext();) {
            String item = (String) i.next();
            out.write(item);
            out.write('\n');
        }

        out.write("</head>\n");
    }

    protected String getPageTitle(PageContentTO page) {
        return "";
    }

    protected void addStyleSheet(Set headerItems, String uri) {
        headerItems.add(getStyleSheetLink(uri));
    }

    private String getStyleSheetLink(String uri) {
        return "<link href=\"" + uri
                + "\" rel=\"stylesheet\" type=\"text/css\">";
    }

    protected void addScript(Set headerItems, String uri) {
        headerItems.add(getScriptElem(uri));
    }

    private String getScriptElem(String uri) {
        return "<script src=\"" + uri
                + "\" type=\"text/javascript\"></script>";
    }

    protected void addFixedPositionCssItems(Set headerItems) {
        addStyleSheet(headerItems, "/dash/fixedPosition.css");
        headerItems.add("<!--[if gte IE 5.5]><![if lt IE 7]>"
                + getStyleSheetLink("/dash/fixedPositionIE.css")
                + "<![endif]><![endif]-->");
    }

    protected void addIEFloatHack(Set headerItems) {
        headerItems.add("<!--[if IE]>"
                + "<style> .IEFloatHack { height: 1% } </style>"
                + "<![endif]-->\n");
    }

    /** Convience routine */
    protected static String esc(String s) {
        return HTMLUtils.escapeEntities(s);
    }

    protected static boolean hasEditPermission() {
        return PermissionsManager.getInstance().hasPermission("pdash.cms");
    }

}
